﻿#pragma TextEncoding="UTF-8"
#pragma rtGlobals=3
#pragma ModuleName=StackTraces
#pragma version=1.3
#pragma DefaultTab = {3,20,4}																	// Set default tab width in Igor Pro 9 and later

// original code and idea by Tony Withers - https://www.wavemetrics.com/user/tony
// rewritten by Stephan Thuermer - https://www.wavemetrics.com/user/chozo

static StrConstant ksIgnoreList = "w_base;MQP_*;"												// list of traces to ignore, wildcards okay
//------------------------------------------------------------------------------------------------

//________________________________________________________________________________________________
//
//							Right-Click Offset Menus for Graph Traces
//________________________________________________________________________________________________

Menu "TracePopup", dynamic
	Submenu "Offset Spectra"
		StackTraces#SaveMenuPosition("y-Offset: Align Traces Here"),/Q, StackTraces#AlignTracesAtXPos(0)	// holding ctrl / command aligns to zero
		StackTraces#SaveMenuPosition("y-Scale: Match Traces Here"), /Q, StackTraces#AlignTracesAtXPos(1)	// holding ctrl / command scales to one
		StackTraces#SaveMenuPosition("y-Order Traces Here"),		/Q, StackTraces#OrderTracesAtXPos()		// holding ctrl / command sorts in descending order
		"Spread Traces Across y Range", /Q, StackTraces#SpreadToFillRange()
		"Spread Traces: Equal Offset", 	/Q, StackTraces#SpreadSpectra(0)
		"Spread Traces: Equal Gap", 	/Q, StackTraces#SpreadSpectra(1)
	End
End

Menu "GraphPopup"
	Submenu "Offset Spectra"
		"Spread Traces Across y Range",	/Q, StackTraces#SpreadToFillRange()
		"Spread Traces: Equal Offset", 	/Q, StackTraces#SpreadSpectra(0)
		"Spread Traces: Equal Gap", 	/Q, StackTraces#SpreadSpectra(1)
	End
End

//################################################################################################

static Function/S SaveMenuPosition(s_menu)														// use to create a menu item in a popup when you need to know where the user clicks to invoke the popup
	String s_menu
	if (WinType("") != 1)																		// don't create variables if Igor is just rebuilding the menu
		return ""
	endif
	GetMouse/W=kwTopWin
	if (V_left<0 || V_top<0)
		return ""
	endif
	if(IgorVersion() < 8.00)																	// compatibility with earlier Igor versions: save mouse position as global variable
		Variable/G V_menuX = V_left, V_menuY = V_top
	endif
	return s_menu
End

//################################################################################################

static Function/S RemoveFromListWC(listStr, zapListStr)											// returns listStr, purged of any items that match an item in ZapListStr.
	String listStr, zapListStr
	String removeStr=""
	Variable i
	for(i = 0; i < ItemsInList(zapListStr); i += 1)
		removeStr += ListMatch(listStr, StringFromList(i, zapListStr))							// Wildcards okay! Case insensitive.
	endfor
	return RemoveFromList(removeStr, listStr, ";", 0)
End

//################################################################################################

static Function getAllTraceOffsets(gName, trace, xOff, yOff, xScale, yScale)
	String gName, trace
	Variable &xOff, &yOff, &xScale, &yScale
	sscanf StringByKey("offset(x)", TraceInfo(gName,trace,0), "="), "{%f,%f}", xOff, yOff
	sscanf StringByKey("muloffset(x)", TraceInfo(gName,trace,0), "="), "{%f,%f}", xScale, yScale
	yScale = yScale == 0? 1 : yScale
	xScale = xScale == 0? 1 : xScale
	return 0
End

//################################################################################################

static Function getTraceOffset(gName, trace, axis)
	String gName, trace
	Variable axis
	Variable xOff = 0, yOff = 0
	sscanf StringByKey("offset(x)", TraceInfo(gName,trace,0), "="), "{%f,%f}", xOff, yOff
	return axis ? yOff : xOff
End

//################################################################################################

static Function getYfromWave(gName, trace, xPos, yVal, yMin)
	String gName, trace
	Variable xPos, &yVal, &yMin
	
	Wave w = TraceNameToWaveRef(gName, trace)
	Wave/Z w_x = XWaveRefFromTrace(gName, trace)
	
	String range
	Variable col2D = 0
	if (WaveDims(w) == 2)																		// handle 2D waterfall plots => extract column
		range = StringByKey("YRANGE", TraceInfo(gName,trace,0))									// extract the last number out of something like "[*][15]" or "[3,50][15]"
		col2D = strsearch(range,"[",Inf,1)
		range = range[col2D+1,inf]
		col2D = str2num(RemoveEnding(range,"]"))
		if (numtype(col2D) == 0)
			Duplicate/R=[][col2D]/FREE w work
			Wave w = work
		else
			return -1
		endif
	endif
	
	yMin = WaveMin(w)
	yVal = NaN
	if(WaveExists(w_x))
		FindLevel/P/Q w_x, xPos
		if (!v_flag && v_levelX > 0 && v_levelX < DimSize(w,0)-1)								// make sure to stay within range
			yVal = w[round(v_levelX)]
		else
			return -1
		endif
	else
		if (x2pnt(w, xPos) > 0 && x2pnt(w, xPos) < DimSize(w,0)-1)
			yVal = w(xPos)
		else
			return -1
		endif
	endif
	return 0
End

//################################################################################################

static Function getMinMaxfromWave(gName, trace, yMin, yMax)										// figure out plotted x-range for trace and finds wave's max and min values in this range
	String gName, trace
	Variable &yMin, &yMax 
	
	wave w = TraceNameToWaveRef(gName, trace)
	wave/Z w_x = XWaveRefFromTrace(gName, trace)
	
	String range
	Variable col2D = 0
	if (WaveDims(w) == 2)
		range = StringByKey("YRANGE", TraceInfo(gName,trace,0))
		col2D = strsearch(range,"[",Inf,1)
		range = range[col2D+1,inf]
		col2D = str2num(RemoveEnding(range,"]"))
		if (numtype(col2D) == 0)
			Duplicate/R=[][col2D]/FREE w work
			Wave w = work
		else
			return -1
		endif
	endif
	
	Variable xOff = 0, yOff = 0, xScale = 0, yScale = 0											// x-offset and y-scale aware
	getAllTraceOffsets(gName, trace, xOff, yOff, xScale, yScale)
	if (WhichListItem("bottom",AxisList(gName)) > -1)
		GetAxis/W=$gName/Q bottom
	elseif (WhichListItem("top",AxisList(gName)) > -1)
		GetAxis/W=$gName/Q top
	else
		return -1
	endif
	
	if(WaveExists(w_x))
		FindLevel /Q w_x, v_min																	// no /P flag here - assumes same x-scaling for w and w_x
		v_min = v_flag ? 0 : V_LevelX
		FindLevel /Q w_x, v_max
		v_max = v_flag ? numpnts(w) : V_LevelX
	endif
	yMax = WaveMax(w, v_min + xOff, v_max + xOff)*yScale
	yMin = WaveMin(w, v_min + xOff, v_max + xOff)*yScale
	return 0
End

//################################################################################################

static Function AlignTracesAtXPos(which)														// use trace popup to y-align traces at clicked x position by setting offsets (which = 0) or multiplicators (which = 1)
	Variable which
	
	GetLastUserMenuInfo																			// figure out graph and trace names
#if(IgorVersion() < 8.00)																		// compatibility with earlier Igor versions
	NVAR V_mouseX = v_menuX
	NVAR V_mouseY = v_menuY
#endif

	String gName = S_graphName, trace = S_traceName
	String s_info = TraceInfo(gName, trace, 0)
	String traces = RemoveFromListWC(TraceNameList("",";",1), ksIgnoreList)						// String traces = RemoveFromListWC(FetchTraces("",traces,1), ksIgnoreList)
	Variable xClick = AxisValFromPixel(gName, StringByKey("XAXIS",s_info) , V_mouseX)
	Variable yClick = AxisValFromPixel(gName, StringByKey("YAXIS",s_info) , V_mouseY)
	
	Variable normVal = (GetKeyState(0) & 2^0)													// holding ctrl / command aligns to zero or scales to one
	yClick = normVal && which == 0 ? 0 : yClick													// shift to zero?
	
	Variable i, err, modVal
	for(i = 0; i < ItemsInList(traces); i += 1)
		trace = StringFromList(i,traces)
		Variable xOff = 0, yOff = 0, xScale = 0, yScale = 0, ywVal, ywMin
		getAllTraceOffsets(gName, trace, xOff, yOff, xScale, yScale)
		err = getYfromWave(gName, trace, xClick-xOff, ywVal, ywMin)								// extract wave min and Y value at X position
		if (err)
			continue
		endif
		if (which)
			modVal = normVal ? 1/(ywVal-ywMin) : (yClick-yOff-ywMin)/(ywVal-ywMin)				// scale to one?
			ModifyGraph muloffset($trace)={,modVal}												// scale to match position
			if (normVal && which)																// if normalized scale offsets down as well (because the scale difference can be huge)
				ModifyGraph offset($trace)={,yOff*modVal}
			endif
		else
			modVal = yClick-ywVal*yScale
			ModifyGraph offset($trace)={,modVal}												// shift to align position
		endif
	endfor
	return 0
End

//################################################################################################

static Function OrderTracesAtXPos()																// reorder traces based on (possibly offset) value at x=xpos
	GetLastUserMenuInfo																			// figure out graph and trace names
#if(IgorVersion() < 8.00)																		// compatibility with earlier Igor versions
	NVAR V_mouseX = v_menuX
#endif

	String gName = S_graphName, trace = S_traceName
	String s_info = TraceInfo(gName, trace, 0)
	String traces = RemoveFromListWC(TraceNameList("",";",1), ksIgnoreList)						// String traces = RemoveFromListWC(FetchTraces("",traces,1), ksIgnoreList)
	Variable xClick = AxisValFromPixel(gName, StringByKey("XAXIS",s_info) , V_mouseX)
	
	Make/Free/T/N=(ItemsInList(traces)) w_traces
	Make/Free/N=(ItemsInList(traces)) w_order
	
	Variable i, err
	for(i = 0 ; i < ItemsInList(traces); i += 1)
		trace = StringFromList(i,traces)
		Variable xOff = 0, yOff = 0, xScale = 0, yScale = 0, ywVal, ywMin
		getAllTraceOffsets(gName, trace, xOff, yOff, xScale, yScale)
		err = getYfromWave(gName, trace, xClick-xOff, ywVal, ywMin)								// extract wave min and Y value at X position
		w_traces[i] = trace
		w_order[i] = err ? yOff : yOff+ywVal*yScale
	endfor
	
	if (GetKeyState(0) & 2^0)																	// holding ctrl / command triggers descending order sorting
		Sort/R w_order, w_order, w_traces
	else
		Sort w_order, w_order, w_traces
	endif
	
	for(i = 0; i < numpnts(w_traces); i += 1)
	#if(IgorVersion() > 7.00)
		ReorderTraces _front_, {$w_traces[i]}
	#endif
	endfor
	return 0
End

//################################################################################################

static Function SpreadToFillRange()																// spread (equally offset) spectra over Y axis range; traces are stacked in the order of plotting
	String traces = RemoveFromListWC(TraceNameList("",";",1), ksIgnoreList)						// String traces = RemoveFromListWC(FetchTraces("",traces,1), ksIgnoreList)
	Variable trNum = ItemsInList(traces)
	Make/Free/T/N=(trNum) w_traces = StringFromList(p,traces)	
	Make/Free/N=(trNum) w_offsets = getTraceOffset("", w_traces[p], 1)

	Variable yMin, yMax, low, high
	getMinMaxfromWave("", w_traces[0], yMin, yMax);			low  = yMin + w_offsets[0]			// find low from first trace
	getMinMaxfromWave("", w_traces[trNum-1], yMin, yMax);	high = yMax + w_offsets[trNum-1]	// find high from last trace

	if (WhichListItem("left",AxisList("")) > -1)
		GetAxis/Q left
	elseif (WhichListItem("right",AxisList("")) > -1)
		GetAxis/Q right
	else
		return -1
	endif
	
	Variable bottomGap = low-v_Min, topGap = v_Max-high
	Variable i, dOffset = (topGap+bottomGap)/(trNum-1)
	for (i = 0; i < trNum; i += 1)																// apply offsets
		ModifyGraph offset($w_traces[i])={,w_offsets[i]- bottomGap + dOffset*i}
	endfor
	return 0
End

//################################################################################################

static Function SpreadSpectra(which)
	Variable which																				// which: 0 = equal offset increment, 1 = equal spacing
	
	String traces = RemoveFromListWC(TraceNameList("",";",1), ksIgnoreList)						// String traces = RemoveFromListWC(FetchTraces("",traces,1), ksIgnoreList)
	Variable trNum = ItemsInList(traces)
	Make/Free/N=(trNum)/T w_traces = StringFromList(p,traces)
	Make/Free/N=(trNum)/Wave w_waves = TraceNameToWaveRef("",w_traces)
	Make/Free/N=(trNum) w_low, w_high, w_range	
	
	Variable i, yMin, yMax
	for (i = 0; i < trNum; i += 1)
		getMinMaxfromWave("", w_traces[i], yMin, yMax);  w_low[i] = yMin;  w_high[i] = yMax;
	endfor
	w_range = w_high-w_low
	
	if (WhichListItem("left",AxisList("")) > -1)
		GetAxis/Q left
	elseif (WhichListItem("right",AxisList("")) > -1)
		GetAxis/Q right
	else
		return -1
	endif
	
	Variable yAxMin = v_min, yAxRange = (v_max-v_min), dataRange = sum(w_range)
	Variable gap = which ? (yAxRange-dataRange)/(trNum-1) : (yAxRange-w_range[trNum-1])/(trNum-1)	// gap or offset between each trace (defined as distance between highest value of spectrum n and lowest value of spectrum n+1)
	Variable newOff
	for (i = 0; i < trNum; i += 1)
		if (i == 0)
			newOff = yAxMin - w_low[0]															// add this to bring first trace down to bottom
		else
			newOff = which ? w_high[i-1] - w_low[i] + gap : w_low[i-1] - w_low[i] + gap
		endif
		ModifyGraph offset($w_traces[i])={,newOff}
		w_high[i] += newOff
		w_low[i]  += newOff
	endfor
	return 0
End

